{-# LANGUAGE DuplicateRecordFields #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE RecordWildCards #-}

module Main where

import Control.Concurrent (forkIO)
import Control.Exception (onException)
import Control.Exception (throwIO)
import Control.Monad (void)
import Data.Acid (openLocalState)
import Data.Acid.Remote (skipAuthenticationCheck, acidServerSockAddr, openRemoteStateSockAddr, skipAuthenticationPerform)
import Data.IMF (Mailbox, parse, mailbox)
import Data.MIME.Charset (defaultCharsets)
import Data.Text.Encoding (encodeUtf8)
import Network.HostName (getHostName)
import Network.Socket (SockAddr(SockAddrUnix))
import Options.Applicative (Parser, ReadM, eitherReader, execParser, strOption, option, hsubparser, command, info, progDesc, long, short, metavar, value, showDefault, help, header, fullDesc, helper, (<**>))
import System.Directory (findExecutable)
import System.Exit (exitWith)
import System.Exit.Codes (codeTempFail)
import System.Posix.User (getRealUserID, userName, getUserEntryForID)
import Web.Scotty (scotty, post, liftAndCatchIO, defaultHandler)

import qualified Data.Text as T
import qualified Data.Text.Lazy.IO as LT

import TiedoteMD.Read
import TiedoteMD.Review
import TiedoteMD.Send
import TiedoteMD.State
import TiedoteMD.Types

serverMain :: SockAddr -> Mailbox -> FilePath -> FilePath -> IO ()
serverMain socket mail repoPath sendmailPath = do
    acid <- openLocalState $ emptyState
    _ <- forkIO $ acidServerSockAddr skipAuthenticationCheck socket acid
    queueMessages acid
    _ <- forkIO $ scotty 3000 $ do
        defaultHandler $ liftAndCatchIO . LT.putStrLn -- Don't return exceptions to clients
        post "/" $ liftAndCatchIO $ void $ forkIO $ updateMessages acid repoPath
    updateMessages acid repoPath
    _ <- forkIO $ manageQueueingMessages acid
    _ <- forkIO $ managePreviews acid mail sendmailPath
    manageQueue acid mail sendmailPath

clientMain :: SockAddr -> Mailbox -> FilePath -> IO ()
clientMain socket mail sendmailPath = do
    acid <- openRemoteStateSockAddr skipAuthenticationPerform socket
        `onException` exitWith codeTempFail
    manageIncomingEmail acid mail sendmailPath

data Config = Config { socketPath :: FilePath, mail :: Mailbox, sendmailCommand :: String, runMode :: RunMode } deriving Show
data RunMode = Server { repoPath :: FilePath } | Recieve deriving Show

config :: Maybe Mailbox -> Parser Config
config mail = Config
    <$> strOption (long "socket" <> short 's' <> metavar "FILE" <> value "/run/tiedote.md/acid.socket" <> showDefault <> help "Path for database socket")
    <*> option readMailbox (long "address" <> short 'a' <> metavar "EMAIL" <> maybe mempty value mail <> help "Email address to send and recieve mail")
    <*> strOption (long "sendmail" <> short 'm' <> metavar "FILE" <> value "sendmail" <> help "Sendmail command")
    <*> hsubparser (command "server" (info serverCommand (progDesc "Run the server"))
        <> command "recieve" (info (pure Recieve) (progDesc "Recieve an email")))

serverCommand :: Parser RunMode
serverCommand = Server <$> strOption (long "repository" <> short 'r' <> metavar "DIRECTORY" <> value "/var/lib/tiedote.md/git-repo" <> showDefault <> help "Path to git repository")

readMailbox :: ReadM Mailbox
readMailbox = eitherReader $ parse (mailbox defaultCharsets) . encodeUtf8 . T.pack

main :: IO ()
main = do
    hostname <- getHostName
    username <- getRealUserID >>= fmap userName . getUserEntryForID
    let defaultMailbox = either (const Nothing) Just $ parse (mailbox defaultCharsets) $
            encodeUtf8 $ T.pack $ username <> "@" <> hostname
    args <- execParser $ info (config defaultMailbox <**> helper) (fullDesc <> header "TiedoteMD mass mailing system")
    sendmailPath <- maybe (throwIO $ FileNotFoundError $ sendmailCommand args) pure =<<
        findExecutable (sendmailCommand args)
    case args of
      Config {runMode = Recieve, ..} -> clientMain (SockAddrUnix socketPath) mail sendmailPath
      Config {runMode = Server {..}, ..} -> serverMain (SockAddrUnix socketPath) mail repoPath sendmailPath
